/*
 ******************************************************************************
 *
 * Copyright Ben Suffolk, 2004/5
 *
 * Ben@vanilla.net
 *
 * You may not reverse engineer, decompile or disassemble the software, except
 * and only to the extent that such activity is expressly permitted by
 * applicable law notwithstanding this limitation.
 *
 * You may not sell, rent, lease, sub-lease or incorporate the whole or any
 * part of the use of the software in any other program or system without the
 * express written agreement
 *
 ******************************************************************************
 *
 * The main Hardware Controller to the USB / DMX Interface
 *
 * When a connection is made to the Interface a new receiver thread is started
 * to process all inbound messages from the Interface. These are then passed
 * out via a NotificationCenter to anybody who might be interested.
 *
 ******************************************************************************
 *
 * History :-
 *
 * Last Author:   $Author: suffo_b $
 * Last modified: $Date: 2005/03/21 09:06:22 $
 * Revision:      $Revision: 1.1 $
 *
 * Author          Date         Description
 *
 * Ben Suffolk     ??/??/2004   Started
 *
 * Ben Suffolk     22/02/2005   Added comments, and tidied up
 *
 ******************************************************************************
*/

#import "HWInterface.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>


/**
 *
 * Class to represent a DMX Value response, and other responses
 *
 **/
@implementation DMXResponse


/**
 *
 * Initilisation method
 *
 **/
- (id)initWithCode:(unsigned char)aCode channel:(unsigned int)aChannel value:(unsigned char)aValue;
 {
 [self init];

 code = aCode;
 channel = aChannel;
 value = aValue;

 return(self);
 }


/**
 *
 * Initilisation method
 *
 **/
- (id)initWithChannel:(unsigned int)aChannel value:(unsigned char)aValue;
 {
 [self init];

 code = RESP_DMXVAL;
 channel = aChannel;
 value = aValue;

 return(self);
 }

/**
 *
 * Methods to get the members variables
 *
 **/
- (unsigned int)getChannel {return(channel);}
- (unsigned char)getValue {return(value);}
- (unsigned char)getCode {return(code);}

@end


/**
 *
 * Class to represent a DMX Version response
 *
 **/
@implementation DMXVersion


/**
 *
 * Initilisation method
 *
 **/
- (id)initWithMajor:(int)aMajor minor:(int)aMinor string:(char *)str
 {
 [self init];

 major = aMajor;
 minor = aMinor;
 
 string = [NSString stringWithCString:str];
 [string retain];

 return(self);
 }


/**
 *
 * Methods to get the members variables
 *
 **/
- (int)getMajor {return(major);}
- (int)getMinor {return(minor);}
- (NSString *)getString {return(string);}

@end



/**
 *
 * The main HW Interface class, see commnets at the top for details of this class
 *
 **/
@implementation HWInterface


/**
 *
 * Get a list of all possible hardware interfaces. As can be passed to initWithPath:
 *
 * @return NSArray of NSSring
 *
 **/
+ (NSArray *)interfacePaths
 {
 NSFileManager *fileManager;
 NSArray *files;
 NSMutableArray *match = [NSMutableArray arrayWithCapacity:5];
 unsigned x;

 /* Get a list of devices */ 
 fileManager = [NSFileManager defaultManager];
 files = [fileManager directoryContentsAtPath:HW_DIR];
 
 /* Loop over the files */
 for(x=0;x<[files count];x++)
  {
  NSString *file = [files objectAtIndex:x];

  /* If the file matches, add it to the array */
  if([file hasPrefix:HW_BASE])
   [match addObject:[NSString stringWithFormat:@"%@%@",HW_DIR,file]];
  }
 
 return([NSArray arrayWithArray:match]);
 }


/**
 *
 * Get a string for the response code
 *
 * @return C style string
 *
 **/
+ (char *)lookupResponse:(unsigned char)aResponse
 {
 
 switch(aResponse)
  {
  case RESP_DMXVAL:
  case RESP_DMXVALH:
   return("DMX Value");
  case RSP_MON_FPS:
   return("DMX Monitor: Frames Per Seconds");
  case RSP_MON_SC:
   return("DMX Monitor: Start Code");
  case RSP_MON_LC:
  case RSP_MON_LCH:
   return("DMX Monitor: Last Code");
  case RESP_VER:
   return("Version");
  case RSP_CH_SET:
   return("Channel Set");
  case RSP_BO_ON:
   return("TX: Blackout On");
  case RSP_BO_OFF:
   return("TX: Blackout Off");
  case RESP_NO_MON:
   return("DMX Monitor: Not started yet");
  case RSP_TX_SLC:
   return("TX: Set Last Channel");
  case RESP_DMX:
   return("RX: Started to receive DMX");
  case RESP_NO_DMX:
   return("RX: Stopped receiving DMX");
  case RESP_NOP:
   return("No Operation");
  case RESP_RSET:
   return("Interface Reset");
  case RESP_RXSC:
   return("RX: Set Start Code");
  case RESP_RXON:
   return("RX: Enabled");
  case RESP_RXOFF:
   return("RX: Disabled");
  case RSP_TX_SSC:
   return("RX: Sest Start Code");
  case RSP_TX_ON:
   return("TX: Enabled");
  case RSP_TX_OFF:
   return("TX: Disabled");
  case RESP_DMXERR:
   return("DMX Framing Error");
  case RESP_HWERR:
   return("Hardware Overflow");
  case RESP_USAVE:
   return("Unknown Command");
  case RSP_NORST:
   return("Reset Failed");
  default:
   return("?");
  }
 
 return(NULL);
 }


/**
 *
 * Open the first DMX interface that can be found. Best really to let the user choose
 *
 * @return The interface, or nil
 *
 **/
- (id)init
 {
 NSArray *interfaces;
 
 /* Get all possible interfaces */
 interfaces = [HWInterface interfacePaths];
 
 /* Check for interfaces */
 if(interfaces == nil || [interfaces count] == 0)
  {
  [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXError" object:@"No Interafces could be found"];
  return(nil);
  }

 /* Open the first one */
 return([self initWithPath:[interfaces objectAtIndex:0]]);
 }


/**
 *
 * Open a specific DMX Interface, and initlise it ready for use.
 *
 * @param aPath The path to the device driver for the Interface to connect with
 *
 **/
- (id)initWithPath:(NSString *)aPath
 {
 int x;

 // Open the device
 if((fd = open([aPath cString],O_RDWR,0)) < 0)
  {
  [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXError" object:[NSString stringWithFormat:@"Unable to open HW at address: %@",aPath]];
  return(nil);
  }
 
 // Keep the path for later (in case we need to reconnect to it)
 path = aPath;
 [path retain];

 // Let the world know we have opened the Interface
 [self postDebug:[NSString stringWithFormat:@"Opened HW at address: %@",path]];
 
 // Initilise the TX & RX universe
 for(x=0;x<512;x++)
  {
  out_universe[x] = 0;
  }

 // Start a receiver thread to process all HW inputs
 threadLock = [[NSLock alloc] init];
 running = YES;
 [NSThread detachNewThreadSelector:@selector(receiverThread:) toTarget:self withObject:self];

 // All done, pass the new HWInterface back
 return(self); 
 }


/**
 *
 * Stop the receiver thread
 *
 **/
- (void)stop
 {
 /* Check if the thread is running */
 if(!running)
  return;
 
 /* Tell the world */
 [self postDebug:@"Stopping receiver thread"];
 running = false;

 /* The lock is realese when the thread stops, so when we can get the lock it must have stopped */
 [threadLock lock];
 [threadLock unlock];

 [self postDebug:@"Receiver thread has stopped"];
 }
 

/**
 *
 * Send a 1 byte command to the interface
 *
 **/
- (void)SendCommand:(unsigned char)aCommand
 {
 int x;
 
 /* If its a TX ON, we need to clear our copy of the universe */
 if(aCommand == CMD_TX_ON)
  {
  for(x=0;x<512;x++)
   out_universe[x] = 0; 
  }

 write(fd,&aCommand,1);
 }


/**
 *
 * Send a command to the interface along with an address.
 *
 * This automaticaly sets the high address bit in the command if needed
 *
 **/
- (void)SendCommand:(unsigned char)aCommand channel:(unsigned int)aChannel
 {
 unsigned char buffer[2];

 /* Check to see if its out of range */
 if(aChannel > 512 || aChannel < 1)
  return;

 // Set code 
 buffer[0] = aCommand;
 
 // Check for high address
 if(((aChannel-1) & 0xFF00))
  buffer[0]++;
 
 buffer[1] = ((aChannel-1) & 0xFF);
 
 write(fd,&buffer,2);
 }


/**
 *
 * Send a command to the interface along with a value.
 *
 **/
- (void)SendCommand:(unsigned char)aCommand value:(unsigned char)aValue
 {
 unsigned char buffer[2];

 buffer[0] = aCommand;
 buffer[1] = aValue;
 write(fd,&buffer,2);
 }


/**
 *
 * Send a command to the interface along with an address and a value.
 *
 * This automaticaly sets the high address bit in the command if needed.
 *
 **/
- (void)SendCommand:(unsigned char)aCommand channel:(unsigned int)aChannel value:(unsigned char)aValue
 {
 unsigned char buffer[3];

 /* Check to see if its out of range */
 if(aChannel > 512 || aChannel < 1)
  return;

 /* Check if its a channel set command, and check the universe if it is*/
 if(aCommand == CMD_CH_SET)
  {
  if(out_universe[aChannel] == aValue)
   return;

  // Record the value
  out_universe[aChannel] = aValue;
  }  
  
 // Set code 
 buffer[0] = aCommand;
 
 // Check for high address
 if(((aChannel-1) & 0xFF00))
  buffer[0]++;
 
 buffer[1] = ((aChannel-1) & 0xFF);
 buffer[2] = aValue;
 
 write(fd,&buffer,3);
 }


/**
 *
 * Receiver thread. This is run in a new thread by the Hardware interface.
 *
 * It keeps running till its told to stop, and all it does it process incomming events 
 * and send Notifications based on them.
 *
 **/
- (void)receiverThread:(id)anObject
 {
 fd_set readfds,writefds,exceptfds;
 struct timeval timeout;
 unsigned char buffer[100],command;
 int count,x,mode;
 unsigned char version;
 char string[100];
 unsigned int address;
 unsigned char value;
 DMXResponse *resp;
 DMXVersion *dongle_ver;

 /* Set up an auto release pool for this thread */
 NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

 /* Use the lock to show the thread is running */
 [threadLock lock];

 /* Mode is used to identify where we are in multi mode responses */
 mode = MODE_CMD;
 string[0] = '\0';

 /* Keep a reference to this thread for the main thread to use */
 receiver = [NSThread currentThread];
 [receiver retain];
 
 // Tell the world we have started
 [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXThreadStart" object:receiver];
 
 
 // Keep going untill its time to stop
 while(running)
  {
  // Set the FD masks
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  FD_SET(fd,&readfds);
  
  // Set the timeout - 1 second
  timeout.tv_sec = 1;
  timeout.tv_usec = 0;

  // Wait for somethig to be received, or a timeout
  select(fd+1,&readfds,&writefds,&exceptfds,&timeout);
  
  // check if something has been received
  if(FD_ISSET(fd, &readfds))
   {
   // Yes, read it in, upto the buffer size
   count = read(fd,buffer,100);

   // Check if there was an error
   if(count < 0)
    {
    // Post the error
    [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXError" object:[NSString stringWithFormat:@"Unable to read data from interface:  %s",strerror(errno)]];
    }
   else
    {
    // Process each character read in the buffer
    for(x=0;x<count;x++)
     {

     // Check which mode we are currently in
     switch(mode)
      {
      // Start of new command
      case MODE_CMD:
      
      // Save the command
      command = buffer[x];

       // Check which command
       switch(command)
        {
        // Multi byte commands
        case RESP_DMXVAL:
        case RSP_MON_FPS:
        case RSP_MON_SC:
        case RSP_MON_LC:
         mode = MODE_ADDR;
         address = 0;
         break;

        // Multi byte high address commands
        case RESP_DMXVALH:
        case RSP_MON_LCH:
         mode = MODE_ADDR;
         address = 0x100;
         break;

        // Version command
        case RESP_VER:
         mode = MODE_VER;
         break;
        
        // Normal commands
        case RSP_CH_SET:
        case RSP_BO_ON:
        case RSP_BO_OFF:
        case RESP_NO_MON:
        case RSP_TX_SLC:
        case RESP_DMX:
        case RESP_NO_DMX:
        case RESP_NOP:
        case RESP_RSET:
        case RESP_RXSC:
        case RESP_RXON:
        case RESP_RXOFF:
        case RSP_TX_SSC:
        case RSP_TX_ON:
        case RSP_TX_OFF:
         resp = [[DMXResponse alloc] initWithCode:command channel:0 value:0];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXResponse" object:resp];
         break;

        // System generated error
        case RESP_DMXERR:
         resp = [[DMXResponse alloc] initWithCode:command channel:0 value:0];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXResponse" object:resp];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXError" object:@"DMX Framing Error"];
         break;

        // System generated error
        case RESP_HWERR:
         resp = [[DMXResponse alloc] initWithCode:command channel:0 value:0];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXResponse" object:resp];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXError" object:@"Hardware Overflow Error"];
         break;
       
        // Unknown command
        default:
         resp = [[DMXResponse alloc] initWithCode:command channel:0 value:0];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXResponse" object:resp];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXError" object:@"Unknown command"];
         break;
        }
       break;
       
      // Read the Version Number Response, and switch to string mode
      case MODE_VER:
       version = buffer[x];
       mode = MODE_STR;
       break;
       
      // Read the Version String Response
      case MODE_STR:
      
       // Still in the string?
       if(buffer[x] != '\0')
        {
        // Keep within a max length of 99 characters
        if(strlen(string) < 99)
         {
         string[strlen(string)+1] = '\0';
         string[strlen(string)] = buffer[x];
         }
        }
       else
        {
        // Send the version response
        dongle_ver = [[DMXVersion alloc] initWithMajor:(version&0xF0)>>4 minor:(version&0x0F) string:string];
        [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXVersion" object:dongle_ver];
        
        // Reset the mode and string for next time
        string[0] = '\0';
        mode = MODE_CMD;
        }
       break;

      // Read the seconds byte of the multi byte command
      case MODE_ADDR:
       address += buffer[x];
       
       switch(command)
        {
        // Got the address, switch mode to get value
        case RESP_DMXVAL:
        case RESP_DMXVALH:
         mode = MODE_VAL;
         break;
        
        // Address data
        case RSP_MON_LC:
        case RSP_MON_LCH:
         resp = [[DMXResponse alloc] initWithCode:command channel:address+1 value:0];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXResponse" object:resp];
         mode = MODE_CMD;
         break;

        // Channel data
        case RSP_MON_FPS:
        case RSP_MON_SC:
         resp = [[DMXResponse alloc] initWithCode:command channel:0 value:address];
         [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXResponse" object:resp];
         mode = MODE_CMD;
         break;
        
        // should never get here
        default:
         mode = MODE_CMD;
         break;
        }
       break;
       
      // Read the incomming DMX Value
      case MODE_VAL:
       value = buffer[x];
       mode = MODE_CMD;
       DMXResponse *resp = [[DMXResponse alloc] initWithChannel:address+1 value:value];
       [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXValue" object:resp];
       break;
      }
     }
    }
   }
  }

 // Tell the world we have stopped
 [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXThreadStop" object:receiver];
 
 // Release the thread copy and the lock
 [receiver release];
 [threadLock unlock];
 }


/**
 *
 * Send a debug string to the notification center so a 
 * logger / console object can see whats going on.
 *
 **/
- (void)postDebug:(NSString *)debug
 {
 [[NSNotificationCenter defaultCenter] postNotificationName:@"DMXDebug" object:debug];
 }


/**
 *
 * Called when the class is finished with
 *
 **/
- (void)dealloc
 {
 
 // If the thread is still running, better stop it
 if(running)
  [self stop];
 
 // Close the interface
 close(fd);
 
 [path release];
 }

@end
